/*
 * bsc1196959_signalfd
 *
 * Fix for CVE-2021-39698, bsc#1196959 (fs/signalfd.c part)
 *
 *  Copyright (c) 2022 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

/* klp-ccp: from fs/signalfd.c */
#include <linux/file.h>
#include <linux/poll.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/signal.h>
#include <linux/list.h>
#include <linux/signalfd.h>

/* klp-ccp: from include/linux/signalfd.h */
#ifdef CONFIG_SIGNALFD

void klpp_signalfd_cleanup(struct sighand_struct *sighand);

#else /* CONFIG_SIGNALFD */
#error "klp-ccp: non-taken branch"
#endif /* CONFIG_SIGNALFD */

/* klp-ccp: from fs/signalfd.c */
#include <linux/compat.h>

void klpp_signalfd_cleanup(struct sighand_struct *sighand)
{
	wait_queue_head_t *wqh = &sighand->signalfd_wqh;
	/*
	 * The lockless check can race with remove_wait_queue() in progress,
	 * but in this case its caller should run under rcu_read_lock() and
	 * sighand_cachep is SLAB_TYPESAFE_BY_RCU, we can safely return.
	 */
	if (likely(!waitqueue_active(wqh)))
		return;

	/* wait_queue_entry_t->func(POLLFREE) should do remove_wait_queue() */
	/*
	 * Fix CVE-2021-39698
	 *  -1 line, +2 lines
	 */
	if (waitqueue_active(wqh))
		__wake_up(wqh, TASK_NORMAL, 0, poll_to_key(EPOLLHUP | POLLFREE));
}
